# combo - content management system
# Copyright (C) 2014  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import urllib
import urlparse

import django
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import logout as auth_logout
from django.contrib.auth import views as auth_views
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db import transaction
from django.http import (Http404, HttpResponse, HttpResponseRedirect,
        HttpResponsePermanentRedirect)
from django.shortcuts import render, resolve_url
from django.template import RequestContext, loader

if django.VERSION >= (1, 8):
    from django.utils import lorem_ipsum
else:
    from django.contrib.webdesign import lorem_ipsum

from django.utils.translation import ugettext as _
from django.forms.widgets import Media

if 'mellon' in settings.INSTALLED_APPS:
    from mellon.utils import get_idps
else:
    get_idps = lambda: []

from combo.data.models import CellBase, Page, ParentContentCell, TextCell
from combo import utils


def login(request, *args, **kwargs):
    if any(get_idps()):
        if not 'next' in request.GET:
            return HttpResponseRedirect(resolve_url('mellon_login'))
        return HttpResponseRedirect(resolve_url('mellon_login') + '?next='
                                    + urllib.quote(request.GET.get('next')))
    return auth_views.login(request, *args, **kwargs)

def logout(request, next_page=None):
    if any(get_idps()):
        return HttpResponseRedirect(resolve_url('mellon_logout'))
    auth_logout(request)
    if next_page is not None:
        next_page = resolve_url(next_page)
    else:
        next_page = '/'
    return HttpResponseRedirect(next_page)

def ajax_page_cell(request, page_pk, cell_reference):
    try:
        page = Page.objects.get(id=page_pk)
    except Page.DoesNotExist:
        raise Http404()
    if not page.is_visible(request.user):
        raise PermissionDenied()
    try:
        cell = CellBase.get_cell(cell_reference, page_id=page_pk)
    except ObjectDoesNotExist:
        raise Http404()

    if not cell.is_visible(request.user):
        raise PermissionDenied()

    context = RequestContext(request, {
        'page': page,
        'request': request,
        'synchronous': True,
        'site_base': request.build_absolute_uri('/')[:-1],
    })

    other_cells = CellBase.get_cells(page_id=page_pk)
    other_cells = [x for x in other_cells if x.is_visible(user=request.user)]
    for other_cell in other_cells:
        if other_cell.get_reference() != cell.get_reference():
            other_cell.modify_global_context(context, request)
        else:
            # Cell can pass data through its own __dict__
            cell.modify_global_context(context, request)

    if django.VERSION < (1, 8, 0):
        return HttpResponse(cell.render(context), content_type='text/html')
    else:
        # FIXME: we bind to an existing template to have the list of
        # context processors applied to the RequestContext; the
        # CellBase.render() method should be extended to accept an additional
        # request parameter, to match the new Template::render() signature.
        template = loader.get_template('combo/page_template.html')
        with context.bind_template(template.template):
            return HttpResponse(cell.render(context), content_type='text/html')


def extend_with_parent_cells(cells):
    for cell in cells[:]:
        if not isinstance(cell, ParentContentCell):
            continue
        idx = cells.index(cell)
        cells[idx:idx+1] = cell.get_cells()


def skeleton(request):
    # Skeleton rendering is used to dynamically produce base templates to use
    # in other applications, based on configured combo cells.
    #
    # It takes a ?source= parameter that should contain the URL we want a
    # template for; it will be used to match the corresponding page, and thus
    # the corresponding content.
    #
    # If there's no matching page, the ?source= parameter will be evaluated
    # against the known services (settings.KNOWN_SERVICES) and an empty page
    # will be created to be used as skeleton.
    #
    # If there was no matching page and no matching service an error will be
    # raised. (403 Access Forbidden)
    #
    # While placeholders holding cells will get their cells rendered, empty
    # placeholders will get themself outputted as template blocks, named
    # placeholder-$name, and with a default content of a block named $name.
    #
    # ex:
    #   {% block placeholder-content %}
    #     {% block content %}
    #     {% endblock %}
    #   {% endblock %}
    if 'source' not in request.GET:
        raise PermissionDenied()
    source = request.GET['source']
    selected_page = None

    # look in redirect pages after the best match for the source
    redirect_pages = Page.objects.exclude(redirect_url__isnull=True).exclude(redirect_url='')
    for page in redirect_pages:
        redirect_url = utils.get_templated_url(page.redirect_url)
        if source.startswith(redirect_url):
            if selected_page is None or len(redirect_url) > len(selected_page.get_redirect_url()):
                selected_page = page

    if selected_page is None:
        # if there was no page found, look for a domain match
        netloc = urlparse.urlparse(source).netloc
        for page in redirect_pages:
            redirect_url = utils.get_templated_url(page.redirect_url)
            if urlparse.urlparse(redirect_url).netloc == netloc:
                selected_page = page
                break

    if selected_page is None:
        # if there's still no page found, look in KNOWN_SERVICES, and
        # return an empty page as template
        for service_id in settings.KNOWN_SERVICES or {}:
            for service_key in settings.KNOWN_SERVICES[service_id]:
                service = settings.KNOWN_SERVICES[service_id][service_key]
                if urlparse.urlparse(service.get('url')).netloc == netloc:
                    selected_page = Page()
                    selected_page.template_name = 'standard'
                    break
            else:
                continue
            break
        else:
            raise PermissionDenied()

    pages = selected_page.get_parents_and_self()
    cells = CellBase.get_cells(page_id=selected_page.id)
    extend_with_parent_cells(cells)

    combo_template = settings.COMBO_PUBLIC_TEMPLATES[selected_page.template_name]

    cell_with_badges = CellBase.get_cells(cell_filter=lambda x: bool(x.get_badge))
    ctx = {
        'page': selected_page,
        'page_cells': cells,
        'pages': pages,
        'request': request,
        'render_skeleton': True,
        'check_badges': bool(cell_with_badges),
        'site_base': request.build_absolute_uri('/')[:-1],
    }

    template_name = combo_template['template']
    response = render(request, template_name, ctx)
    response['X-Combo-Page-Id'] = str(selected_page.id)
    return response


class StyleDemoPage(object):
    def __init__(self, template_name):
        self.template_name = template_name or 'standard'

    def __enter__(self):
        self.page = Page(public=False, title=_('Style Demo'))
        self.page.template_name = self.template_name
        self.page.save()
        combo_template = settings.COMBO_PUBLIC_TEMPLATES[self.page.template_name]
        for i, placeholder in enumerate(combo_template['placeholders'].keys()):
            cell = TextCell(page=self.page, placeholder=placeholder, order=0)
            if placeholder == 'footer':
                cell.text = '<p>%s</p>' % lorem_ipsum.sentence()
            else:
                cell.text = '<h2>%s</h2>%s' % (
                        lorem_ipsum.words(3, common=False),
                        '\n'.join(['<p>%s</p>' % x for x in lorem_ipsum.paragraphs(3)]))
            cell.save()

            if i == 0:
                # add a template selector in first placeholder
                cell = TextCell(page=self.page, placeholder=placeholder, order=1)
                options_html = []
                for template_key, template_dict in settings.COMBO_PUBLIC_TEMPLATES.items():
                    attr = 'value="%s"' % template_key
                    if template_key == self.page.template_name:
                        attr += ' selected="selected"'
                    options_html.append(u'<option %s>%s</option>' % (attr, template_dict['name']))
                cell.text = u'''<form><select name="template">%s</select>
                                <button>%s</button></form>''' % (
                                        '\n'.join(options_html),
                                        _('Select'))
                cell.save()

        return self.page

    def __exit__(self, type, value, traceback):
        self.page.delete()

def style(request):
    if not settings.TEMPLATE_DEBUG:
        raise Http404()
    messages.info(request, _('Info notice'))
    messages.warning(request, _('Warning notice'))
    messages.error(request, _('Error notice'))

    with transaction.atomic():
        with StyleDemoPage(request.GET.get('template')) as page:
            return publish_page(request, page)


def empty_site(request):
    return render(request, 'combo/empty_site.html', {})


def page(request):
    url = request.path_info
    parts = [x for x in request.path_info.strip('/').split('/') if x]
    if len(parts) == 1 and parts[0] == 'index':
        return HttpResponsePermanentRedirect('/')
    if not parts:
        parts = ['index']
    try:
        page = Page.objects.get(slug=parts[-1])
    except Page.DoesNotExist:
        if Page.objects.count() == 0 and parts == ['index']:
            return empty_site(request)
        else:
            if not url.endswith('/') and settings.APPEND_SLASH:
                # this is useful to allow /login, /manage, and other non-page
                # URLs to work.
                return HttpResponsePermanentRedirect(url + '/')
            raise Http404()

    if page.get_online_url() != url:
        if not url.endswith('/') and settings.APPEND_SLASH:
            return HttpResponsePermanentRedirect(url + '/')
        raise Http404()

    return publish_page(request, page)

def publish_page(request, page, status=200, template_name=None):
    pages = page.get_parents_and_self()

    if not page.is_visible(request.user):
        if not request.user.is_authenticated():
            from django.contrib.auth.views import redirect_to_login
            return redirect_to_login(request.path)
        raise PermissionDenied()

    if page.redirect_url:
        return HttpResponseRedirect(page.get_redirect_url())

    cells = CellBase.get_cells(page_id=page.id)
    extend_with_parent_cells(cells)
    cells = [x for x in cells if x.is_visible(user=request.user)]
    cell_with_badges = CellBase.get_cells(cell_filter=lambda x: bool(x.get_badge))

    ctx = {
        'check_badges': bool(cell_with_badges),
        'page': page,
        'page_cells': cells,
        'pages': pages,
        'request': request,
        'media': sum((cell.media for cell in cells), Media())
    }

    for cell in cells:
        cell.modify_global_context(ctx, request)

    if not template_name:
        combo_template = settings.COMBO_PUBLIC_TEMPLATES[page.template_name]
        template_name = combo_template['template']

    return render(request, template_name, ctx, status=status)

def error404(request, *args, **kwargs):
    if not hasattr(request, 'user'):
        # this happens when the 404 handler is called early on, for example
        # when the given hostname doesn't exist as a tenant
        from django.views.defaults import page_not_found
        return page_not_found(request)
    page = Page()
    page.template_name = 'standard'
    return publish_page(request, page, status=404, template_name='combo/404.html')


def menu_badges(request):
    context = RequestContext(request, {'request': request})
    page_ids = request.GET.getlist('page[]')
    cells = CellBase.get_cells(
            cell_filter=lambda x: bool(x.get_badge),
            page_id__in=page_ids)
    badges = {}
    for cell in cells:
        if cell.page_id in badges:
            continue
        badge = cell.get_badge(context)
        if badge:
            badges[cell.page_id] = badge
    return HttpResponse(json.dumps(badges))
